__all__ = ["CallableCondition"]

from typing import (
    TYPE_CHECKING,
    Callable,
    TypeAlias,
)


if TYPE_CHECKING:
    from clsx.abc.conditions.condition import Condition


CallableCondition: TypeAlias = Callable[[], "Condition"]



__all__ = [
    "BoolCondition",
    "CallableCondition",
    "CallableExpression",
    "ClassName",
    "ClassNameCharacterIterable",
    "ClassNameCharacterStream",
    "ClassNameIterable",
    "Condition",
    "ConditionalExpressionTuple",
    "DictExpression",
    "Expression",
    "ExpressionIterable",
    "ExpressionList",
    "ExpressionSet",
    "ExpressionTuple",
    "HashableExpression",
    "NestedClassNameCharacterIterable",
    "UnhashableExpression",
    "VariadicExpressionTuple",
]

from clsx.abc.class_names import (
    ClassName,
    ClassNameCharacterIterable,
    ClassNameCharacterStream,
    ClassNameIterable,
    NestedClassNameCharacterIterable,
)
from clsx.abc.conditions import (
    BoolCondition,
    CallableCondition,
    Condition,
)
from clsx.abc.expressions import (
    CallableExpression,
    ConditionalExpressionTuple,
    DictExpression,
    Expression,
    ExpressionIterable,
    ExpressionList,
    ExpressionSet,
    ExpressionTuple,
    HashableExpression,
    UnhashableExpression,
    VariadicExpressionTuple,
)



__all__ = ["ExpressionEvaluatorTest"]

from itertools import chain
from typing import cast
from unittest import TestCase

from clsx.abc import (
    CallableExpression,
    ClassName,
    DictExpression,
    Expression,
    ExpressionList,
    ExpressionSet,
)
from clsx.contrib.itertools import dedup
from clsx.contrib.stdlib import RepresentableLambda
from clsx.evaluation.expression_evaluator import ExpressionEvaluator


class ExpressionEvaluatorTest(TestCase):
    BLANK = str()

    SPACE = " "

    CLASS_NAMES_LIST = ["nice", "nice--better"]

    CLASS_NAMES_STR = str.join(SPACE, CLASS_NAMES_LIST)

    WHITESPACES_LIST = [SPACE, "\f", "\n", "\r", "\t", "\v"]

    WHITESPACES_STR = str.join("", WHITESPACES_LIST)

    STR_DATA: list[tuple[ClassName, str]] = [
        (BLANK, BLANK),
        (CLASS_NAMES_STR, CLASS_NAMES_STR),
        (f" {CLASS_NAMES_STR} {CLASS_NAMES_STR} ", CLASS_NAMES_STR),
        (WHITESPACES_STR, BLANK),
    ]

    LIST_DATA: list[tuple[ExpressionList, str]] = [
        ([expression, expression, expression, expression], expected)
        for expression, expected in STR_DATA
    ]

    SET_DATA: list[tuple[ExpressionSet, str]] = [
        (set(expression), expected) for expression, expected in LIST_DATA
    ]

    LAZY_DATA: list[tuple[CallableExpression, str]] = [
        (cast(CallableExpression, RepresentableLambda(expression)), expected)
        for expression, expected in chain(
            STR_DATA,
            LIST_DATA,
        )
    ]

    DICT_DATA: list[tuple[DictExpression, str]] = list(
        chain(
            map(
                lambda kv: ({kv[0]: True}, kv[1]),
                chain(
                    STR_DATA,
                    LAZY_DATA,
                ),
            ),
            map(
                lambda kv: ({kv[0]: False}, ""),
                chain(
                    STR_DATA,
                    LAZY_DATA,
                ),
            ),
        ),
    )

    def evaluate(
        self,
        expression: Expression,
    ) -> str:
        return str.join(
            self.SPACE,
            dedup(
                ExpressionEvaluator.evaluate(
                    expression,
                ),
            ),
        )

    def test_evaluating_str(self) -> None:
        for expression, expected in self.STR_DATA:
            with self.subTest(expression=expression, expected=expected):
                self.assertEqual(
                    self.evaluate(expression),
                    expected,
                )

    def test_evaluating_list(self) -> None:
        for expression, expected in self.LIST_DATA:
            with self.subTest(expression=expression, expected=expected):
                self.assertEqual(
                    self.evaluate(expression),
                    expected,
                )

    def test_evaluating_set(self) -> None:
        for expression, expected in self.SET_DATA:
            with self.subTest(expression=expression, expected=expected):
                self.assertSetEqual(
                    set(self.evaluate(expression).split(self.SPACE)),
                    set(expected.split(self.SPACE)),
                )

    def test_evaluating_lazy(self) -> None:
        for expression, expected in self.LAZY_DATA:
            with self.subTest(expression=expression, expected=expected):
                self.assertEqual(
                    self.evaluate(expression),
                    expected,
                )

    def test_evaluating_dict(self) -> None:
        for expression, expected in self.DICT_DATA:
            with self.subTest(expression=expression, expected=expected):
                self.assertEqual(
                    self.evaluate(expression),
                    expected,
                )



__all__ = ["Expression"]

from typing import (
    TypeAlias,
    Union,
)

from clsx.abc.expressions.hashable_expression import HashableExpression
from clsx.abc.expressions.unhashable_expression import UnhashableExpression


Expression: TypeAlias = Union[
    HashableExpression,
    UnhashableExpression,
]



__all__ = ["ClassNameCharacterStream"]

from typing import (
    TypeAlias,
    Union,
)

from clsx.abc.class_names.class_name_character_iterable import (
    ClassNameCharacterIterable,
)
from clsx.abc.class_names.nested_class_name_character_iterable import (
    NestedClassNameCharacterIterable,
)


ClassNameCharacterStream: TypeAlias = Union[
    ClassNameCharacterIterable,
    NestedClassNameCharacterIterable,
]



__all__ = ["ExpressionTuple"]

from typing import (
    TypeAlias,
    Union,
)

from clsx.abc.expressions.conditional_expression_tuple import ConditionalExpressionTuple
from clsx.abc.expressions.variadic_expression_tuple import VariadicExpressionTuple


ExpressionTuple: TypeAlias = Union[
    ConditionalExpressionTuple,
    VariadicExpressionTuple,
]



__all__ = ["ClassNameParser"]

from clsx.abc import (
    ClassNameCharacterStream,
    ClassNameIterable,
)
from clsx.contrib.itertools import flatten_str


class ClassNameParser:
    @staticmethod
    def parse(
        stream: ClassNameCharacterStream,
    ) -> ClassNameIterable:
        buffer: list[str] = []
        for character in flatten_str(stream):
            if character.isspace():
                if buffer:
                    yield str.join("", buffer)
                    buffer = []
            else:
                buffer.append(character)
        if buffer:
            yield str.join("", buffer)



__all__ = ["clsx"]

from typing import (
    Any,
    overload,
)

from clsx.abc import (
    ClassName,
    Condition,
    Expression,
)
from clsx.contrib.itertools import dedup
from clsx.evaluation import ExpressionEvaluator


@overload
def clsx(
    *expressions: Expression,
) -> ClassName: ...


@overload
def clsx(
    expression: Expression,
    condition: Condition,
    /,
) -> ClassName: ...


def clsx(
    *expressions: Any,
) -> ClassName:
    return " ".join(
        dedup(
            ExpressionEvaluator.evaluate(
                expressions,
            ),
        ),
    )



__all__ = [
    "ClassName",
    "ClassNameCharacterIterable",
    "ClassNameCharacterStream",
    "ClassNameIterable",
    "NestedClassNameCharacterIterable",
]

from clsx.abc.class_names.class_name import ClassName
from clsx.abc.class_names.class_name_character_iterable import (
    ClassNameCharacterIterable,
)
from clsx.abc.class_names.class_name_character_stream import ClassNameCharacterStream
from clsx.abc.class_names.class_name_iterable import ClassNameIterable
from clsx.abc.class_names.nested_class_name_character_iterable import (
    NestedClassNameCharacterIterable,
)



__all__ = ["ConditionalExpressionTuple"]

from typing import (
    TYPE_CHECKING,
    TypeAlias,
)

from clsx.abc.conditions.condition import Condition


if TYPE_CHECKING:
    from clsx.abc.expressions.expression import Expression


ConditionalExpressionTuple: TypeAlias = tuple["Expression", Condition]



__all__ = ["HashableExpression"]

from typing import (
    TypeAlias,
    Union,
)

from clsx.abc.class_names.class_name import ClassName
from clsx.abc.expressions.callable_expression import CallableExpression
from clsx.abc.expressions.expression_iterable import ExpressionIterable
from clsx.abc.expressions.expression_tuple import ExpressionTuple


HashableExpression: TypeAlias = Union[
    None,
    ClassName,
    ExpressionTuple,
    ExpressionIterable,
    CallableExpression,
]



__all__ = ["flatten_str"]

from typing import Iterable

from clsx.contrib.itertools.abc import NestedStrIterable


def flatten_str(iterable: NestedStrIterable) -> Iterable[str]:
    for value in iterable:
        if isinstance(value, str) and len(value) == 1:
            yield value
        else:
            yield from flatten_str(value)



__all__ = [
    "CallableExpression",
    "ConditionalExpressionTuple",
    "DictExpression",
    "Expression",
    "ExpressionIterable",
    "ExpressionList",
    "ExpressionSet",
    "ExpressionTuple",
    "HashableExpression",
    "UnhashableExpression",
    "VariadicExpressionTuple",
]


from clsx.abc.expressions.callable_expression import CallableExpression
from clsx.abc.expressions.conditional_expression_tuple import ConditionalExpressionTuple
from clsx.abc.expressions.dict_expression import DictExpression
from clsx.abc.expressions.expression import Expression
from clsx.abc.expressions.expression_iterable import ExpressionIterable
from clsx.abc.expressions.expression_list import ExpressionList
from clsx.abc.expressions.expression_set import ExpressionSet
from clsx.abc.expressions.expression_tuple import ExpressionTuple
from clsx.abc.expressions.hashable_expression import HashableExpression
from clsx.abc.expressions.unhashable_expression import UnhashableExpression
from clsx.abc.expressions.variadic_expression_tuple import VariadicExpressionTuple



__all__ = ["ExpressionEvaluator"]

from typing import (
    Iterable,
    cast,
)

from clsx.abc import (
    BoolCondition,
    CallableCondition,
    CallableExpression,
    ClassNameIterable,
    Condition,
    ConditionalExpressionTuple,
    DictExpression,
    Expression,
    ExpressionIterable,
    ExpressionTuple,
    VariadicExpressionTuple,
)
from clsx.parsing import ClassNameParser


class ExpressionEvaluator:
    @staticmethod
    def evaluate(
        expression: Expression,
    ) -> ClassNameIterable:
        yield from ExpressionEvaluator.evaluate_dynamic(expression)

    @staticmethod
    def evaluate_dynamic(
        expression: Expression,
    ) -> ClassNameIterable:
        if expression is None:
            yield from []
        elif isinstance(expression, str):
            yield from ExpressionEvaluator.evaluate_str(expression)
        elif isinstance(expression, tuple):
            yield from ExpressionEvaluator.evaluate_tuple(expression)
        elif isinstance(expression, dict):
            yield from ExpressionEvaluator.evaluate_dict(expression)
        elif isinstance(expression, list):
            yield from ExpressionEvaluator.evaluate_iterable(expression)
        elif isinstance(expression, set):
            yield from ExpressionEvaluator.evaluate_iterable(expression)
        elif isinstance(expression, Iterable):
            yield from ExpressionEvaluator.evaluate_iterable(
                cast(
                    ExpressionIterable,
                    expression,
                ),
            )
        elif callable(expression):
            yield from ExpressionEvaluator.evaluate_callable(expression)

    @staticmethod
    def evaluate_str(
        expression: str,
    ) -> ClassNameIterable:
        yield from ClassNameParser.parse(expression)

    @staticmethod
    def evaluate_tuple(
        expression: ExpressionTuple,
    ) -> ClassNameIterable:
        if len(expression) == 2:
            yield from ExpressionEvaluator.evaluate_conditional_expression_tuple(
                cast(
                    ConditionalExpressionTuple,
                    expression,
                ),
            )
        else:
            yield from ExpressionEvaluator.evaluate_variadic_expression_tuple(
                cast(
                    VariadicExpressionTuple,
                    expression,
                ),
            )

    @staticmethod
    def evaluate_conditional_expression_tuple(
        expression: ConditionalExpressionTuple,
    ) -> ClassNameIterable:
        sub_expression, condition = expression
        flag = ExpressionEvaluator.evaluate_maybe_condition(condition)
        is_flag_bool = isinstance(flag, bool)
        if not is_flag_bool or flag:
            yield from ExpressionEvaluator.evaluate_dynamic(sub_expression)
            if not is_flag_bool:
                yield from ExpressionEvaluator.evaluate_dynamic(
                    cast(
                        Expression,
                        flag,
                    ),
                )

    @staticmethod
    def evaluate_variadic_expression_tuple(
        expression: VariadicExpressionTuple,
    ) -> ClassNameIterable:
        yield from ExpressionEvaluator.evaluate_iterable(expression)

    @staticmethod
    def evaluate_dict(
        expression: DictExpression,
    ) -> ClassNameIterable:
        for key, value in expression.items():
            yield from ExpressionEvaluator.evaluate_tuple((key, value))

    @staticmethod
    def evaluate_iterable(
        expression: ExpressionIterable,
    ) -> ClassNameIterable:
        for sub_expression in expression:
            yield from ExpressionEvaluator.evaluate_dynamic(sub_expression)

    @staticmethod
    def evaluate_callable(
        expression: CallableExpression,
    ) -> ClassNameIterable:
        yield from ExpressionEvaluator.evaluate_dynamic(expression())

    @staticmethod
    def evaluate_maybe_condition(
        maybe_condition: Condition | Expression,
    ) -> BoolCondition | Expression:
        if callable(maybe_condition):
            return ExpressionEvaluator.evaluate_maybe_condition(maybe_condition())
        elif isinstance(maybe_condition, bool):
            return ExpressionEvaluator.evaluate_bool_condition(maybe_condition)
        return maybe_condition

    @staticmethod
    def evaluate_condition(
        condition: Condition,
    ) -> bool:
        if callable(condition):
            return ExpressionEvaluator.evaluate_lazy_condition(condition)
        return ExpressionEvaluator.evaluate_bool_condition(condition)

    @staticmethod
    def evaluate_lazy_condition(
        condition: CallableCondition,
    ) -> bool:
        return ExpressionEvaluator.evaluate_condition(condition())

    @staticmethod
    def evaluate_bool_condition(
        condition: BoolCondition,
    ) -> bool:
        return not not condition



__all__ = ["ExpressionIterable"]

from typing import (
    TYPE_CHECKING,
    Iterable,
    TypeAlias,
)


if TYPE_CHECKING:
    from clsx.abc.expressions.expression import Expression


ExpressionIterable: TypeAlias = Iterable["Expression"]



__all__ = ["RepresentableLambda"]

from pprint import pformat
from typing import (
    Generic,
    TypeVar,
)


T = TypeVar("T")


class RepresentableLambda(Generic[T]):
    def __init__(
        self,
        value: T,
    ) -> None:
        self.value = value

    def __call__(self) -> T:
        return self.value

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({pformat(self.value)})"



__all__ = ["CallableExpression"]

from typing import (
    TYPE_CHECKING,
    Callable,
    TypeAlias,
)


if TYPE_CHECKING:
    from clsx.abc.expressions.expression import Expression


CallableExpression: TypeAlias = Callable[[], "Expression"]



__all__ = ["DictExpression"]

from typing import TypeAlias

from clsx.abc.conditions.condition import Condition
from clsx.abc.expressions.hashable_expression import HashableExpression


DictExpression: TypeAlias = dict[HashableExpression, Condition]



__all__ = ["Condition"]

from typing import (
    TypeAlias,
    Union,
)

from clsx.abc.conditions.bool_condition import BoolCondition
from clsx.abc.conditions.callable_condition import CallableCondition


Condition: TypeAlias = Union[
    BoolCondition,
    CallableCondition,
]



__all__ = ["NestedStrIterable"]

from typing import (
    Iterable,
    TypeAlias,
    Union,
)


NestedStrIterable: TypeAlias = Union[
    str,
    Iterable[str],
    Iterable["NestedStrIterable"],
]



__all__ = ["NestedClassNameCharacterIterable"]

from typing import TypeAlias

from clsx.contrib.itertools.abc import NestedStrIterable


NestedClassNameCharacterIterable: TypeAlias = NestedStrIterable



__all__ = ["dedup"]

from typing import (
    Iterable,
    TypeVar,
)


T = TypeVar("T")


def dedup(iterable: Iterable[T]) -> Iterable[T]:
    seen = set()
    for value in iterable:
        if value not in seen:
            yield value
            seen.add(value)



__all__ = [
    "BoolCondition",
    "CallableCondition",
    "Condition",
]

from clsx.abc.conditions.bool_condition import BoolCondition
from clsx.abc.conditions.callable_condition import CallableCondition
from clsx.abc.conditions.condition import Condition



__all__ = ["ClsxTest"]

from unittest import TestCase

from clsx.api._clsx import clsx


class ClsxTest(TestCase):
    def test_readme_usage_section_examples(self) -> None:
        self.assertEqual(clsx(), "")
        self.assertEqual(clsx(""), "")
        self.assertEqual(clsx("foo"), "foo")
        self.assertEqual(clsx("foo foo"), "foo")
        self.assertEqual(clsx("foo \n\t foo"), "foo")

        self.assertEqual(clsx("foo", None), "foo")
        self.assertEqual(clsx("foo", None, "bar"), "foo bar")

        self.assertEqual(clsx("foo", None, "bar", None), "foo bar")
        self.assertEqual(clsx("foo", None, "bar", None, "baz"), "foo bar baz")

        self.assertEqual(clsx("foo", False), "")
        self.assertEqual(clsx("foo", True), "foo")

        self.assertEqual(clsx("foo", "bar"), "foo bar")
        self.assertEqual(clsx("foo", "bar", "baz"), "foo bar baz")
        self.assertEqual(clsx("foo", "bar", "baz", "qux"), "foo bar baz qux")

        self.assertEqual(clsx(["foo", "bar"]), "foo bar")
        self.assertEqual(clsx(["foo", "bar"], ["baz", "qux"]), "foo bar baz qux")

        self.assertEqual(clsx([("foo", True), ("foo", "bar", "baz")]), "foo bar baz")

        self.assertEqual(clsx(["foo foo"], "foo", "bar"), "foo bar")
        self.assertEqual(clsx("foo", ["foo", "bar", "baz"]), "foo bar baz")

        self.assertEqual(clsx({"foo": True, "bar": False}), "foo")
        self.assertEqual(clsx({"foo": True, "bar": True}), "foo bar")
        self.assertEqual(clsx({(lambda: "foo"): True}), "foo")

        self.assertEqual(clsx(lambda: [lambda: ("foo", True)]), "foo")



__all__ = ["UnhashableExpression"]

from typing import (
    TYPE_CHECKING,
    TypeAlias,
    Union,
)

from clsx.abc.expressions.expression_list import ExpressionList
from clsx.abc.expressions.expression_set import ExpressionSet


if TYPE_CHECKING:
    from clsx.abc.expressions.dict_expression import DictExpression


UnhashableExpression: TypeAlias = Union[
    "DictExpression",
    ExpressionList,
    ExpressionSet,
]



__all__ = ["VariadicExpressionTuple"]

from typing import (
    TYPE_CHECKING,
    TypeAlias,
)


if TYPE_CHECKING:
    from clsx.abc.expressions.expression import Expression


VariadicExpressionTuple: TypeAlias = tuple["Expression", ...]



__all__ = ["main"]

import operator
import sys

from contextlib import suppress
from pathlib import Path
from typing import (
    Iterable,
    ParamSpec,
    TextIO,
    TypeVar,
    cast,
)


P = ParamSpec("P")

R = TypeVar("R")

T = TypeVar("T")


WORK_DIR = Path.cwd()

VENV_DIR = WORK_DIR / ".venv"


def read(buffer: TextIO) -> str:
    return buffer.read()


def get_export_expression_span(
    source: str,
) -> tuple[tuple[int, int, int], tuple[int, int, int]]:
    line_no = 1
    column_no = 0
    start_line_no = -1
    start_column_no = -1
    start_offset = -1
    end_line_no = -1
    end_column_no = -1
    end_offset = -1
    lookbehind: list[tuple[tuple[int, int, int], str]] = []
    prefix = "__all__=["
    prefix_chars = list(prefix)
    for char_index, char in cast(Iterable[tuple[int, str]], enumerate(source)):
        if char == "\n":
            line_no += 1
            column_no = 0
        else:
            column_no += 1
        if start_offset == -1:
            if not (char.isspace() or char == "\\"):
                lookbehind.append(((line_no, column_no, char_index), char))
                if len(lookbehind) > len(prefix_chars):
                    lookbehind.pop(0)
            if len(lookbehind) > 0:
                lookbehind_value = list(map(operator.itemgetter(1), lookbehind))
                found_prefix = lookbehind_value == prefix_chars
                is_prefix_at_start = lookbehind[0][0][1] == 1
                if found_prefix and is_prefix_at_start:
                    lookbehind.clear()
                    start_line_no = line_no
                    start_column_no = column_no
                    start_offset = char_index
        if start_offset != -1 and char == "]":
            end_line_no = line_no
            end_column_no = column_no
            end_offset = char_index
            break
    start = start_line_no, start_column_no, start_offset
    end = end_line_no, end_column_no, end_offset
    return start, end


def collect_existing_exported_symbols(source: str) -> set[tuple[int, str]]:
    existing_exported_symbols: set[tuple[int, str]] = set()
    start, end = get_export_expression_span(source)
    if start[2] == -1 or end[2] == -1:
        return existing_exported_symbols
    lines = source[start[2] + 1 : end[2]].split("\n")
    for line_no, line in enumerate(lines):
        for raw_string in line.split(","):
            symbol_line_no = start[0] + line_no
            symbol_name = raw_string.strip()[1:-1].strip()
            if not symbol_name:
                continue
            symbol = (symbol_line_no, symbol_name)
            existing_exported_symbols.add(symbol)
    return existing_exported_symbols


def collect_imported_symbols(source: str) -> set[tuple[int, str]]:
    imported_symbols: set[tuple[int, str]] = set()
    lines = list(map(lambda line: line.strip(), source.split("\n")))
    for line_no, line in enumerate(lines):
        if line.startswith("from"):
            parts = line.split(" ", maxsplit=3)
            if len(parts) != 4:
                continue
            if parts[2] != "import":
                continue
            if parts[-1] == "(":
                index_of_tuple_end = line_no + lines[line_no:].index(")")
                section = "\n".join(lines[line_no + 1 : index_of_tuple_end])
                symbol_names = [
                    symbol_name.strip()
                    for symbol_name in section.split(",")
                    if symbol_name.strip()
                ]
                for symbol_name_index, symbol_name in enumerate(symbol_names):
                    symbol = (line_no + symbol_name_index + 1, symbol_name)
                    imported_symbols.add(symbol)
            elif "," in parts[-1]:
                symbol_names = [
                    symbol_name.strip()
                    for symbol_name in parts[-1].split(",")
                    if symbol_name.strip()
                ]
                for symbol_name_index, symbol_name in enumerate(symbol_names):
                    symbol = (line_no + symbol_name_index + 1, symbol_name)
                    imported_symbols.add(symbol)
            else:
                symbol = (line_no, parts[-1])
                imported_symbols.add(symbol)
    return imported_symbols


def replace_exported_symbols(
    source: str,
    start: tuple[int, int, int],
    end: tuple[int, int, int],
    symbols: set[tuple[int, str]],
) -> str:
    exported_symbols = sorted(set(map(operator.itemgetter(1), symbols)))
    prefix = ""
    suffix = ""
    indent = ""
    if len(exported_symbols) > 1:
        prefix = "\n"
        suffix = ",\n"
        indent = "    "
    inner = ",\n".join(
        [f'{indent}"{symbol}"' for symbol in exported_symbols],
    )
    rendered = f"{prefix}{inner}{suffix}"
    source_chars = list(source)
    source_chars[start[2] + 1 : end[2]] = list(rendered)
    code = "".join(source_chars)
    return code


def write(
    buffer: TextIO,
    output: str,
) -> None:
    buffer.write(output)
    buffer.flush()


def process_stdin(argv: list[str]) -> int:
    source = read(sys.stdin)
    start, end = get_export_expression_span(source)
    if -1 in (start[2], end[2]):
        write(sys.stdout, source)
        return -1
    existing_exported_symbols = collect_existing_exported_symbols(source)
    imported_symbols: set[tuple[int, str]]
    if "-export-all" in argv:
        imported_symbols = collect_imported_symbols(source)
    else:
        imported_symbols = set()
    output = replace_exported_symbols(
        source,
        start,
        end,
        existing_exported_symbols.union(
            imported_symbols,
        ),
    )
    try:
        write(sys.stdout, output)
    except Exception as error:
        write(sys.stderr, f"{str(error)}\n")
        with suppress(Exception):
            write(sys.stdout, source)
        return 1
    return 0


def process_source_file(source_file_path: Path) -> int:
    source = source_file_path.read_text()
    start, end = get_export_expression_span(source)
    if -1 in (start[2], end[2]):
        return -1
    existing_exported_symbols = collect_existing_exported_symbols(source)
    imported_symbols: set[tuple[int, str]]
    if source_file_path.name == "__init__.py":
        imported_symbols = collect_imported_symbols(source)
    else:
        imported_symbols = set()
    output = replace_exported_symbols(
        source,
        start,
        end,
        existing_exported_symbols.union(
            imported_symbols,
        ),
    )
    try:
        with source_file_path.open("w") as source_file:
            write(source_file, output)
    except Exception as error:
        write(sys.stderr, f"{str(error)}\n")
        with suppress(Exception):
            with source_file_path.open("w") as source_file:
                write(source_file, source)
        return 1
    return 0


def process_source_files() -> int:
    for source_file_path in WORK_DIR.rglob("*.py"):
        if str(source_file_path).startswith(str(VENV_DIR)):
            continue
        status = process_source_file(source_file_path)
        if status == -1:
            write(
                sys.stderr,
                f"{source_file_path}: No exported py symbols found, skipping\n",
            )
        elif status != 0:
            write(
                sys.stderr,
                f"{source_file_path}: Failed to format exported py symbols\n",
            )
            return status
        write(
            sys.stderr,
            f"{source_file_path}: Formatted exported py symbols\n",
        )
    return 0


def main(
    argv: list[str] | None = None,
) -> int:
    if argv is None:
        argv = sys.argv.copy()
    if "-stdin" in argv:
        return process_stdin(argv)
    return process_source_files()


if __name__ == "__main__":
    sys.exit(main(argv=sys.argv.copy()))
